;;!emacs
;;
;; LCD-ENTRY:    The OO-Browser|Bob Weiner|bwcto@users.sf.net|Multi-Language Object-oriented Browser|03-Jan-2002|04.08|http://www.sf.net/projects/oo-browser
;;
;; FILE:         br-vers.el
;; SUMMARY:      OO-Browser revision number as `br-version' variable.
;; USAGE:        GNU Emacs Lisp Library
;; KEYWORDS:     oop, tools
;;
;; AUTHOR:       Bob Weiner
;;
;; ORIG-DATE:    16-Mar-95 at 12:31:05
;; LAST-MOD:      3-Jan-02 at 17:07:12 by Bob Weiner
;;
;; Copyright (C) 2000, 2001, 2002  Bob Weiner
;; Copyright (C) 1995, 1997, 1998, 1999  BeOpen.com
;; See the file BR-COPY for license information.
;;
;; This file is part of the OO-Browser.

;;; ************************************************************************
;;; Public variables
;;; ************************************************************************

(defconst br-version "04.08" "The OO-Browser revision number.")
